@@ -16,14 +16,14 @@ limitations under the License.
 
 """
 
-import inspect
-import traceback
+import re
 
+from falcon import api_helpers as helpers
 from falcon.request import Request
 from falcon.response import Response
 import falcon.responders
 from falcon.status_codes import HTTP_416
-from falcon import api_helpers as helpers
+from falcon import util
 
 from falcon.http_error import HTTPError
 from falcon import DEFAULT_MEDIA_TYPE
@@ -37,7 +37,8 @@ class API(object):
 
     """
 
-    __slots__ = ('_after', '_before', '_media_type', '_routes')
+    __slots__ = ('_after', '_before', '_error_handlers', '_media_type',
+                 '_routes', '_default_route', '_sinks')
 
     def __init__(self, media_type=DEFAULT_MEDIA_TYPE, before=None, after=None):
         """Initialize a new Falcon API instances
@@ -57,11 +58,15 @@ class API(object):
         """
 
         self._routes = []
+        self._sinks = []
+        self._default_route = None
         self._media_type = media_type
 
         self._before = helpers.prepare_global_hooks(before)
         self._after = helpers.prepare_global_hooks(after)
 
+        self._error_handlers = []
+
     def __call__(self, env, start_response):
         """WSGI "app" method
 
@@ -69,19 +74,41 @@ class API(object):
 
         Args:
             env: A WSGI environment dictionary
-            start_response: A WSGI helper method for setting status and headers
-                on a response.
+            start_response: A WSGI helper method for setting status and
+                headers on a response.
 
         """
 
         req = Request(env)
         resp = Response()
 
-        responder, params, na_responder = self._get_responder(
+        responder, params = self._get_responder(
             req.path, req.method)
 
         try:
-            responder(req, resp, **params)
+            # NOTE(kgriffs): Using an inner try..except in order to
+            # address the case when err_handler raises HTTPError.
+            #
+            # NOTE(kgriffs): Coverage is giving false negatives,
+            # so disabled on relevant lines. All paths are tested
+            # afaict.
+            try:
+                responder(req, resp, **params)  # pragma: no cover
+            except Exception as ex:
+                for err_type, err_handler in self._error_handlers:
+                    if isinstance(ex, err_type):
+                        err_handler(ex, req, resp, params)
+                        break  # pragma: no cover
+
+                else:
+                    # PERF(kgriffs): This will propagate HTTPError to
+                    # the handler below. It makes handling HTTPError
+                    # less efficient, but that is OK since error cases
+                    # don't need to be as fast as the happy path, and
+                    # indeed, should perhaps be slower to create
+                    # backpressure on clients that are issuing bad
+                    # requests.
+                    raise
 
         except HTTPError as ex:
             resp.status = ex.status
@@ -91,42 +118,6 @@ class API(object):
             if req.client_accepts('application/json'):
                 resp.body = ex.json()
 
-        except TypeError as ex:
-            # NOTE(kgriffs): Get the stack trace up here since we can just
-            # use this convenience function which graps the last raised
-            # exception context.
-            stack_trace = traceback.format_exc()
-
-            # See if the method doesn't support the given route's params, to
-            # support assigning multiple routes to the same resource.
-            try:
-                argspec = responder.wrapped_argspec
-            except AttributeError:
-                argspec = inspect.getargspec(responder)
-
-            # First three args should be (self, req, resp)
-            if argspec.args[0] == 'self':
-                offset = 3
-            else:
-                offset = 2
-
-            args_needed = set(argspec.args[offset:])
-            args_given = set(params.keys())
-
-            # Reset the response
-            resp = Response()
-
-            # Does the responder require more or fewer args than given?
-            if args_needed != args_given:
-                req.log_error('A responder method could not be found with the '
-                              'correct arguments.')
-                na_responder(req, resp)
-            else:
-                # Error caused by something else
-                req.log_error('A responder method (on_*) raised TypeError. %s'
-                              % stack_trace)
-                falcon.responders.internal_server_error(req, resp)
-
         #
         # Set status and headers
         #
@@ -159,7 +150,7 @@ class API(object):
 
         A resource is an instance of a class that defines various on_*
         "responder" methods, one for each HTTP method the resource
-        allows. For example, to support GET, simply define an "on_get"
+        allows. For example, to support GET, simply define an `on_get`
         responder. If a client requests an unsupported method, Falcon
         will respond with "405 Method not allowed".
 
@@ -170,7 +161,7 @@ class API(object):
                 pass
 
         In addition, if the route's uri template contains field
-        expressions, any responders that desires to receive requests
+        expressions, any responder that desires to receive requests
         for that route must accept arguments named after the respective
         field names defined in the template. For example, given the
         following uri template:
@@ -187,15 +178,11 @@ class API(object):
             def on_put(self, req, resp):
                 pass
 
-        Falcon would respond to the client's request with "405 Method
-        not allowed." This allows you to define multiple routes to the
-        same resource, e.g., in order to support GET for "/widget/1234"
-        and POST to "/widgets". In this last example, a POST to
-        "/widget/5000" would result in a 405 response.
-
         Args:
             uri_template: Relative URI template. Currently only Level 1
-                templates are supported. See also RFC 6570.
+                templates are supported. See also RFC 6570. Care must be
+                taken to ensure the template does not mask any sink
+                patterns (see also add_sink).
             resource: Object which represents an HTTP/REST "resource". Falcon
                 will pass "GET" requests to on_get, "PUT" requests to on_put,
                 etc. If any HTTP methods are not supported by your resource,
@@ -205,16 +192,103 @@ class API(object):
         """
 
         uri_fields, path_template = helpers.compile_uri_template(uri_template)
-        method_map, na_responder = helpers.create_http_method_map(
+        method_map = helpers.create_http_method_map(
             resource, uri_fields, self._before, self._after)
 
         # Insert at the head of the list in case we get duplicate
         # adds (will cause the last one to win).
-        self._routes.insert(0, (path_template, method_map, na_responder))
+        self._routes.insert(0, (path_template, method_map))
 
-#----------------------------------------------------------------------------
-# Helpers
-#----------------------------------------------------------------------------
+    def add_sink(self, sink, prefix=r'/'):
+        """Add a "sink" responder to the API.
+
+        If no route matches a request, but the path in the requested URI
+        matches the specified prefix, Falcon will pass control to the
+        given sink, regardless of the HTTP method requested.
+
+        Args:
+            sink: A callable of the form:
+
+                 func(req, resp)
+
+            prefix: A regex string, typically starting with '/', which
+                will trigger the sink if it matches the path portion of the
+                request's URI. Both strings and precompiled regex objects
+                may be specified. Characters are matched starting at the
+                beginning of the URI path.
+
+                Named groups are converted to kwargs and passed to
+                the sink as such.
+
+                If the route collides with a route's URI template, the
+                route will mask the sink (see also add_route).
+
+        """
+
+        if not hasattr(prefix, 'match'):
+            # Assume it is a string
+            prefix = re.compile(prefix)
+
+        # NOTE(kgriffs): Insert at the head of the list such that
+        # in the case of a duplicate prefix, the last one added
+        # is preferred.
+        self._sinks.insert(0, (prefix, sink))
+
+    # TODO(kgriffs): Remove this functionality in Falcon version 0.2.0
+    @util.deprecated('Please migrate to add_sink(...) ASAP.')
+    def set_default_route(self, default_resource):
+        """DEPRECATED: Route all the unrouted requests to a default resource
+
+        NOTE: If a default route is defined, all sinks are ignored.
+
+        Args:
+            default_resource: Object which works like an HTTP/REST resource.
+                Falcon will pass "GET" requests to on_get, "PUT" requests to
+                on_put, etc. If you want to exclude some HTTP method from the
+                default routing, just simply don't define the corresponding
+                request handlers.
+
+        """
+
+        self._default_route = helpers.create_http_method_map(
+            default_resource, set(), self._before, self._after)
+
+    def add_error_handler(self, exception, handler=None):
+        """Adds a handler for a given exception type
+
+        Args:
+            exception: Whenever an exception occurs when handling a request
+                that is an instance of this exception class, the given
+                handler callable will be used to handle the exception.
+            handler: Callable that gets called with (ex, req, resp, params)
+                when there is a matching exception when handling a
+                request. If not specified, the handler will default to
+                exception.handle, in which case the method is expected to
+                be static (i.e., decorated with @staticmethod) and take
+                the same params described above.
+
+                Note: A handler can either raise an instance of HTTPError
+                or modify resp manually in order to communicate information
+                about the issue to the client.
+
+        """
+
+        if handler is None:
+            try:
+                handler = exception.handle
+            except AttributeError:
+                raise AttributeError('handler must either be specified '
+                                     'explicitly or defined as a static'
+                                     'method named "handle" that is a '
+                                     'member of the given exception class.')
+
+        # Insert at the head of the list in case we get duplicate
+        # adds (will cause the last one to win).
+        self._error_handlers.insert(0, (exception, handler))
+
+    # ----------------------------------------------------------------------
+    # Helpers
+    # ----------------------------------------------------------------------
 
     def _get_responder(self, path, method):
         """Searches routes for a matching responder
@@ -222,15 +296,17 @@ class API(object):
         Args:
             path: URI path to search (without query string)
             method: HTTP method (uppercase) requested
+
         Returns:
-            A 2-member tuple, containing a responder callable and a dict
-            containing parsed path fields, if any were specified in
-            the matching route's URI template
+            A 3-member tuple, containing a responder callable, a dict
+            containing parsed path fields (if any were specified in
+            the matching route's URI template), and a reference to
+            the "method not allowed" responder for the resource.
 
         """
 
         for route in self._routes:
-            path_template, method_map, na_responder = route
+            path_template, method_map = route
             m = path_template.match(path)
             if m:
                 params = m.groupdict()
@@ -242,8 +318,26 @@ class API(object):
 
                 break
         else:
-            responder = falcon.responders.path_not_found
             params = {}
-            na_responder = falcon.responders.create_method_not_allowed([])
 
-        return (responder, params, na_responder)
+            if self._default_route is None:
+
+                for pattern, sink in self._sinks:
+                    m = pattern.match(path)
+                    if m:
+                        params = m.groupdict()
+                        responder = sink
+
+                        break
+                else:
+                    responder = falcon.responders.path_not_found
+
+            else:
+                method_map = self._default_route
+
+                try:
+                    responder = method_map[method]
+                except KeyError:
+                    responder = falcon.responders.bad_request
+
+        return (responder, params)
