@@ -20,11 +20,13 @@ import json
 import sys
 
 if sys.version_info < (2, 7):  # pragma: no cover
-    from ordereddict import OrderedDict
-else:
+    # NOTE(kgriffs): We could use the module from PyPI, but ordering isn't
+    # critical in JSON, and Falcon eschews dependencies.
+    OrderedDict = dict
+else:  # pragma: no cover
     from collections import OrderedDict
 
-from falcon.status_codes import *
+from falcon import util
 
 
 class HTTPError(Exception):
@@ -73,7 +75,7 @@ class HTTPError(Exception):
             headers: A dictionary of extra headers to return in the
                 response to the client (default None).
             href: A URL someone can visit to find out more information
-                (default None).
+                (default None). Unicode characters are percent-encoded.
             href_text: If href is given, use this as the friendly
                 title/description for the link (defaults to "API documentation
                 for this error").
@@ -92,7 +94,7 @@ class HTTPError(Exception):
         if href:
             link = self.link = OrderedDict()
             link['text'] = (href_text or 'API documention for this error')
-            link['href'] = href
+            link['href'] = util.percent_escape(href)
             link['rel'] = 'help'
         else:
             self.link = None
@@ -124,4 +126,5 @@ class HTTPError(Exception):
         if self.link:
             obj['link'] = self.link
 
-        return json.dumps(obj, indent=4, separators=(',', ': '))
+        return json.dumps(obj, indent=4, separators=(',', ': '),
+                          ensure_ascii=False)
