@@ -18,14 +18,29 @@ limitations under the License.
 
 from datetime import datetime
 
+try:
+    # NOTE(kgrifs): In Python 2.6 and 2.7, socket._fileobject is a
+    # standard way of exposing a socket as a file-like object, and
+    # is used by wsgiref for wsgi.input.
+    import socket
+    NativeStream = socket._fileobject
+except AttributeError:  # pragma nocover
+    # NOTE(kgriffs): In Python 3.3, wsgiref implements wsgi.input
+    # using _io.BufferedReader which is an alias of io.BufferedReader
+    import io
+    NativeStream = io.BufferedReader
+
+import mimeparse
 import six
 
 from falcon.exceptions import HTTPBadRequest
 from falcon import util
+from falcon.util import uri
 from falcon import request_helpers as helpers
 
+
 DEFAULT_ERROR_LOG_FORMAT = (u'{0:%Y-%m-%d %H:%M:%S} [FALCON] [ERROR]'
-                            u' {1} {2}?{3} => ')
+                            u' {1} {2}{3} => ')
 
 TRUE_STRINGS = ('true', 'True', 'yes')
 FALSE_STRINGS = ('false', 'False', 'no')
@@ -56,8 +71,10 @@ class Request(object):
     """
 
     __slots__ = (
+        '_cached_headers',
+        '_cached_uri',
+        '_cached_relative_uri',
         'env',
-        '_headers',
         'method',
         '_params',
         'path',
@@ -73,15 +90,13 @@ class Request(object):
 
         Args:
             env: A WSGI environment dict passed in from the server. See also
-                the PEP-333 spec.
+                the PEP-3333 spec.
 
         """
-
         self.env = env
 
         self._wsgierrors = env['wsgi.errors']
         self.stream = env['wsgi.input']
-
         self.method = env['REQUEST_METHOD']
 
         # Normalize path
@@ -96,19 +111,42 @@ class Request(object):
 
         # QUERY_STRING isn't required to be in env, so let's check
         # PERF: if...in is faster than using env.get(...)
-        if 'QUERY_STRING' in env:
-            self.query_string = env['QUERY_STRING']
+        if 'QUERY_STRING' in env and env['QUERY_STRING']:
+
+            # TODO(kgriffs): Should this escape individual values instead
+            # of the entire string? The way it is now, this:
+            #
+            #   x=ab%2Bcd%3D42%2C9
+            #
+            # becomes this:
+            #
+            #   x=ab+cd=42,9
+            #
+            self.query_string = uri.decode(env['QUERY_STRING'])
+
         else:
-            self.query_string = ''
+            self.query_string = six.text_type()
 
         # PERF: Don't parse it if we don't have to!
         if self.query_string:
-            self._params = helpers.parse_query_string(self.query_string)
+            self._params = uri.parse_query_string(self.query_string)
         else:
             self._params = {}
 
-        self._headers = helpers.parse_headers(env)
+        helpers.normalize_headers(env)
+        self._cached_headers = {}
 
+        self._cached_uri = None
+        self._cached_relative_uri = None
+
+        # NOTE(kgriffs): Wrap wsgi.input if needed to make read() more robust,
+        # normalizing semantics between, e.g., gunicorn and wsgiref.
+        if isinstance(self.stream, NativeStream):  # pragma: nocover
+            # NOTE(kgriffs): coverage can't detect that this *is* actually
+            # covered since the test that does so uses multiprocessing.
+            self.stream = helpers.Body(self.stream, self.content_length)
+
+    # TODO(kgriffs): Use the nocover pragma only for the six.PY3 if..else
     def log_error(self, message):  # pragma: no cover
         """Log an error to wsgi.error
 
@@ -122,9 +160,15 @@ class Request(object):
 
         """
 
+        if self.query_string:
+            query_string_formatted = '?' + self.query_string
+        else:
+            query_string_formatted = ''
+
         log_line = (
             DEFAULT_ERROR_LOG_FORMAT.
-            format(datetime.now(), self.method, self.path, self.query_string)
+            format(datetime.now(), self.method, self.path,
+                   query_string_formatted)
         )
 
         if six.PY3:
@@ -139,23 +183,68 @@ class Request(object):
     @property
     def client_accepts_json(self):
         """Return True if the Accept header indicates JSON support."""
-
-        accept = self._get_header_by_wsgi_name('ACCEPT')
-        return ((accept is not None) and
-                (('application/json' in accept) or ('*/*' in accept)))
+        return self.client_accepts('application/json')
 
     @property
     def client_accepts_xml(self):
         """Return True if the Accept header indicates XML support."""
+        return self.client_accepts('application/xml')
 
-        accept = self._get_header_by_wsgi_name('ACCEPT')
-        return ((accept is not None) and
-                (('application/xml' in accept) or ('*/*' in accept)))
+    def client_accepts(self, media_type):
+        """Returns the client's preferred media type.
+
+        Args:
+            media_type: Media type to check
+
+        Returns:
+            True IFF the client has indicated in the Accept header that
+            they accept at least one of the specified media types.
+        """
+
+        accept = self.accept
+
+        # PERF(kgriffs): Usually the following will be true, so
+        # try it first.
+        if (accept == media_type) or (accept == '*/*'):
+            return True
+
+        # Fall back to full-blown parsing
+        try:
+            return mimeparse.quality(media_type, accept) != 0.0
+        except ValueError:
+            return False
+
+    def client_prefers(self, media_types):
+        """Returns the client's preferred media type given several choices.
+
+        Args:
+            media_types: One or more media types from which to choose the
+                client's preferred type. This value MUST be an iterable
+                collection of strings.
+
+        Returns:
+            The client's preferred media type, based on the Accept header,
+            or None if the client does not accept any of the specified
+            types.
+        """
+
+        try:
+            # NOTE(kgriffs): best_match will return '' if no match is found
+            preferred_type = mimeparse.best_match(media_types, self.accept)
+        except ValueError:
+            # Value for the accept header was not formatted correctly
+            preferred_type = ''
+
+        return (preferred_type if preferred_type else None)
 
     @property
     def accept(self):
-        """Value of the Accept header, or None if not found."""
-        return self._get_header_by_wsgi_name('ACCEPT')
+        """Value of the Accept header, or */* if not found per RFC."""
+        accept = self._get_header_by_wsgi_name('HTTP_ACCEPT')
+
+        # NOTE(kgriffs): Per RFC, missing accept header is
+        # equivalent to '*/*'
+        return '*/*' if accept is None else accept
 
     @property
     def app(self):
@@ -165,7 +254,7 @@ class Request(object):
     @property
     def auth(self):
         """Value of the Authorization header, or None if not found."""
-        return self._get_header_by_wsgi_name('AUTHORIZATION')
+        return self._get_header_by_wsgi_name('HTTP_AUTHORIZATION')
 
     @property
     def content_length(self):
@@ -178,7 +267,7 @@ class Request(object):
             HTTPBadRequest: The header had a value, but it wasn't
                 formatted correctly or was a negative number.
         """
-        value = self._get_header_by_wsgi_name('CONTENT_LENGTH')
+        value = self._get_header_by_wsgi_name('HTTP_CONTENT_LENGTH')
         if value:
             try:
                 value_as_int = int(value)
@@ -194,12 +283,12 @@ class Request(object):
             else:
                 return value_as_int
 
-        # implicit return None
+        return None
 
     @property
     def content_type(self):
         """Value of the Content-Type header, or None if not found."""
-        return self._get_header_by_wsgi_name('CONTENT_TYPE')
+        return self._get_header_by_wsgi_name('HTTP_CONTENT_TYPE')
 
     @property
     def date(self):
@@ -216,7 +305,7 @@ class Request(object):
 
         """
 
-        http_date = self._get_header_by_wsgi_name('DATE')
+        http_date = self._get_header_by_wsgi_name('HTTP_DATE')
         try:
             return util.http_date_to_dt(http_date)
         except ValueError:
@@ -227,32 +316,32 @@ class Request(object):
     @property
     def expect(self):
         """Value of the Expect header, or None if missing."""
-        return self._get_header_by_wsgi_name('EXPECT')
+        return self._get_header_by_wsgi_name('HTTP_EXPECT')
 
     @property
     def if_match(self):
         """Value of the If-Match header, or None if missing."""
-        return self._get_header_by_wsgi_name('IF_MATCH')
+        return self._get_header_by_wsgi_name('HTTP_IF_MATCH')
 
     @property
     def if_none_match(self):
         """Value of the If-None-Match header, or None if missing."""
-        return self._get_header_by_wsgi_name('IF_NONE_MATCH')
+        return self._get_header_by_wsgi_name('HTTP_IF_NONE_MATCH')
 
     @property
     def if_modified_since(self):
         """Value of the If-Modified-Since header, or None if missing."""
-        return self._get_header_by_wsgi_name('IF_MODIFIED_SINCE')
+        return self._get_header_by_wsgi_name('HTTP_IF_MODIFIED_SINCE')
 
     @property
     def if_unmodified_since(self):
         """Value of the If-Unmodified-Since header, or None if missing."""
-        return self._get_header_by_wsgi_name('IF_UNMODIFIED_SINCE')
+        return self._get_header_by_wsgi_name('HTTP_IF_UNMODIFIED_SINCE')
 
     @property
     def if_range(self):
         """Value of the If-Range header, or None if missing."""
-        return self._get_header_by_wsgi_name('IF_RANGE')
+        return self._get_header_by_wsgi_name('HTTP_IF_RANGE')
 
     @property
     def protocol(self):
@@ -279,7 +368,7 @@ class Request(object):
                 formatted correctly.
         """
 
-        value = self._get_header_by_wsgi_name('RANGE')
+        value = self._get_header_by_wsgi_name('HTTP_RANGE')
 
         if value:
             if ',' in value:
@@ -305,22 +394,27 @@ class Request(object):
                                               href=href,
                                               href_text=href_text)
 
-     # implicit return None
+        return None
 
     @property
     def uri(self):
         """The fully-qualified URI for the request."""
 
-        # PERF: For small numbers of items, '+' is faster than ''.join(...)
-        value = (self.protocol + '://' +
-                 self.get_header('host') +
-                 self.app +
-                 self.path)
+        if self._cached_uri is None:
+            # PERF: For small numbers of items, '+' is faster
+            # than ''.join(...). Concatenation is also generally
+            # faster than formatting.
+            value = (self.protocol + '://' +
+                     self.get_header('host') +
+                     self.app +
+                     self.path)
 
-        if self.query_string:
-            value = value + '?' + self.query_string
+            if self.query_string:
+                value = value + '?' + self.query_string
 
-        return value
+            self._cached_uri = value
+
+        return self._cached_uri
 
     url = uri
     """Alias for uri"""
@@ -328,15 +422,49 @@ class Request(object):
     @property
     def relative_uri(self):
         """The path + query string portion of the full URI."""
-        if self.query_string:
-            return self.app + self.path + '?' + self.query_string
 
-        return self.app + self.path
+        if self._cached_relative_uri is None:
+            if self.query_string:
+                self._cached_relative_uri = (self.app + self.path + '?' +
+                                             self.query_string)
+            else:
+                self._cached_relative_uri = self.app + self.path
+
+        return self._cached_relative_uri
 
     @property
     def user_agent(self):
         """Value of the User-Agent string, or None if missing."""
-        return self._get_header_by_wsgi_name('USER_AGENT')
+        return self._get_header_by_wsgi_name('HTTP_USER_AGENT')
+
+    @property
+    def headers(self):
+        """Get raw HTTP headers
+
+        Build a temporary dictionary of dash-separated HTTP headers,
+        which can be used as a whole, like, to perform an HTTP request.
+
+        If you want to lookup a header, please use `get_header` instead.
+
+        Returns:
+            A new dictionary of HTTP headers.
+
+        """
+
+        # NOTE(kgriffs: First time here will cache the dict so all we
+        # have to do is clone it in the future.
+        if not self._cached_headers:
+            headers = self._cached_headers
+
+            env = self.env
+            for name, value in env.items():
+                if name.startswith('HTTP_'):
+                    # NOTE(kgriffs): Don't take the time to fix the case
+                    # since headers are supposed to be case-insensitive
+                    # anyway.
+                    headers[name[5:].replace('_', '-')] = value
+
+        return self._cached_headers.copy()
 
     def get_header(self, name, required=False):
         """Return a header value as a string
@@ -362,7 +490,7 @@ class Request(object):
             # Don't take the time to cache beforehand, using HTTP naming.
             # This will be faster, assuming that most headers are looked
             # up only once, and not all headers will be requested.
-            return self._headers[name.upper().replace('-', '_')]
+            return self.env['HTTP_' + name.upper().replace('-', '_')]
         except KeyError:
             if not required:
                 return None
@@ -390,13 +518,15 @@ class Request(object):
 
         """
 
+        params = self._params
+
         # PERF: Use if..in since it is a good all-around performer; we don't
         #       know how likely params are to be specified by clients.
-        if name in self._params:
+        if name in params:
             if store is not None:
-                store[name] = self._params[name]
+                store[name] = params[name]
 
-            return self._params[name]
+            return params[name]
 
         if not required:
             return None
@@ -434,10 +564,12 @@ class Request(object):
 
         """
 
+        params = self._params
+
         # PERF: Use if..in since it is a good all-around performer; we don't
         #       know how likely params are to be specified by clients.
-        if name in self._params:
-            val = self._params[name]
+        if name in params:
+            val = params[name]
             try:
                 val = int(val)
             except ValueError:
@@ -493,10 +625,12 @@ class Request(object):
 
         """
 
+        params = self._params
+
         # PERF: Use if..in since it is a good all-around performer; we don't
         #       know how likely params are to be specified by clients.
-        if name in self._params:
-            val = self._params[name]
+        if name in params:
+            val = params[name]
             if val in TRUE_STRINGS:
                 val = True
             elif val in FALSE_STRINGS:
@@ -527,7 +661,9 @@ class Request(object):
             name: Parameter name, case-sensitive (e.g., 'limit')
             transform: An optional transform function that takes as input
                 each element in the list as a string and outputs a transformed
-                element for inclusion in the list that will be returned.
+                element for inclusion in the list that will be returned. For
+                example, passing the int function will transform list items
+                into numbers.
             required: Set to True to raise HTTPBadRequest instead of returning
                 gracefully when the parameter is not found or is not an
                 integer (default False)
@@ -536,21 +672,45 @@ class Request(object):
 
         Returns:
             The value of the param if it is found. Otherwise, returns None
-            unless required is True.
+            unless required is True. for partial lists, None will be returned
+            as a placeholder. For example:
+
+                things=1,,3
+
+            would be returned as:
+
+                ['1', None, '3']
+
+            while this:
+
+                things=,,,
+
+            would just be retured as:
+
+                [None, None, None, None]
 
         Raises
             HTTPBadRequest: The param was not found in the request, but was
                 required.
-
         """
 
+        params = self._params
+
         # PERF: Use if..in since it is a good all-around performer; we don't
         #       know how likely params are to be specified by clients.
-        if name in self._params:
-            items = self._params[name].split(',')
-            if transform is not None:
+        if name in params:
+            items = params[name].split(',')
+
+            # PERF(kgriffs): Use if-else rather than a DRY approach
+            # that sets transform to a passthrough function; avoids
+            # function calling overhead.
+            if transform is None:
+                items = [i if i != '' else None
+                         for i in items]
+            else:
                 try:
-                    items = [transform(x) for x in items]
+                    items = [transform(i) if i != '' else None
+                             for i in items]
                 except ValueError:
                     desc = ('The value of the "' + name + '" query parameter '
                             'is not formatted correctly.')
@@ -583,6 +743,6 @@ class Request(object):
 
         """
         try:
-            return self._headers[name] or None
+            return self.env[name] or None
         except KeyError:
             return None
